;DUCK
;ATtiny88/ATmega88/168/328 @ 8 MHz (internal R/C osc.)
;(C) Arne Rossius, http://www.ebps.de.vu/
;
;FUSES (ATtiny88):
;  low  = 0xE2 (8 MHz int. R/C, BOD disabled, slow start-up)
;  high = 0xDF (defaults)
;
;PORTS
;  PB0   = Push-button (to ground)
;  PB1-2 = Speaker PWM signals [OC1A, OC1B] (only PB1 used on prototype)
;  PB3   = Always-on input (active low, jumper on ISP header) [ISP]
;  PB4-5 = [unused, ISP]
;  PB6-7 = [unused]
;  PC0-5 = LED matrix columns 0~5 (active high)
;  PD0-5 = LED matrix rows 0~5 (active high, external NPN transistors)
;  PD6   = [unused]
;  PD7   = Audio amplifier enable (active low, protype only)
;
;VERSIONS
;  2015-12-15  Arne Rossius
;   * first version
;  2016-12-04  Arne Rossius
;   * added double buffering for display
;   * synchronized delays with display refresh rate
;   * increased display refresh rate from 82 Hz to 100 Hz
;  2019-05-01  Arne Rossius
;   * modified for ATmega88/168/328 and ATtiny88 from original ATmega8 version
;   * use WDT interrupt mode instead of reset mode to save power

;.include "m88def.inc"
.include "tn88def.inc"

.def	sreg_backup = R2
.def	repeat = R3
.def	leds = R4 ;used in mux interrupt
.def	spl_backup = R5
.def	sph_backup = R6
.def	pwm_timer = R7
.def	delay = R8
.def	new_delay = R9

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	inttmp = R19 ;used in interrupts
.def	intcnt = R20 ;used in interrupts
.def	pwm = R21 ;used in mux interrupt
.def	mux = R22 ;used in mux interrupt
.def	button = R23 ;used in mux interrupt
.def	wakeup_count = R24
.def	flags = R25
	.equ	fButtonShort = 0
	.equ	fButtonLong = 1
	.equ	fExpectWdr = 2
	.equ	fUpdateLEDs = 3
	.equ	fDelay = 4
;R27:R26 used for X pointer
;R29:R28 used for Y pointer
;R31:R30 used for Z pointer

.equ	BUTTON_SHORT_TIME = 2 ;20 ms
.equ	BUTTON_LONG_TIME = 100 ;1 s

;distance between any 2 values: >= 2
;distance between PWM_6 and PWM_7: >= 6
.equ	PWM_1 = 2
.equ	PWM_2 = 5
.equ	PWM_3 = 11
.equ	PWM_4 = 25
.equ	PWM_5 = 51
.equ	PWM_6 = 103
.equ	PWM_7 = 208 ;8 MHz / 64 / 6 / 208 = ~100.16 Hz

;===============================================================================

.dseg
	RAM_FrameBackup: .byte 36 ;virtual matrix size is 6x6 LEDs
	RAM_FrameShadow: .byte 36
	RAM_FrameLED:    .byte 36

;===============================================================================

;ATtiny88 has simplified timer 0
.ifndef TCCR0B
  .equ TCCR0B = TCCR0A
.endif

;auto-select "in" or "lds" depending on address
.macro _in
  .if (@1 >= 0x60)
	lds	@0, @1
  .else
	in	@0, @1
  .endif
.endmacro

;auto-select "out" or "sts" depending on address
.macro _out
  .if (@0 >= 0x60)
	sts	@0, @1
  .else
	out	@0, @1
  .endif
.endmacro

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	
.org WDTaddr
	;watchdog interrupt: used for wake-up
	reti
	
.org OVF1addr
	;Timer 1 overflow (update PWM values for audio output): 31.25 kHz
	rjmp	ovf1

.org OVF0addr
	;Timer 0 overflow (LED PWM/Mux): interval according to current PWM step
	;interrupt duration (worst cases):
	;   no mux restart:             max. 117 cycles (< 2 timer ticks)
	;   mux restart, no LED update: max. 135 cycles (< 3 timer ticks)
	;   mux restart + LED update:   max. 325 cycles (< 6 timer ticks)
	in	sreg_backup, SREG
	
	;set LEDs
	ldi	inttmp, 0x00
	out	PORTC, inttmp ;LEDs off
	in	inttmp, PORTD
	andi	inttmp, 0x80 ;keep audio amp enable pin unchanged
	or	inttmp, mux
	out	PORTD, inttmp ;switch row
	;delay for transistors to switch
	out	TCNT0, pwm_timer ;set timer for next interrupt
	push	ZL
	push	ZH
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	out	PORTC, leds ;LEDs on
	
	;max. 26 cycles from timer overflow to here
	
	;LED PWM
	subi	pwm, 1
	brcc	ovf0_end ;+3 cycles if this path taken
	;restart PWM sequence
	ldi	pwm, 6
	
	;LED Multiplexing
	lsl	mux
	sbrs	mux, 6
	rjmp	ovf0_end ;+7 cycles if this path taken
	;restart multiplexing sequence
	ldi	mux, 0x01
	
	;reset watchdog (~100 Hz)
	wdr
	
	;+8 cycles to here
	
	;read button (~100 Hz) - max. 10 cycles
	sbic	PINB, 0
	rjmp	ovf0_button_released
	inc	button ;button pressed
	cpi	button, BUTTON_LONG_TIME
	brlo	ovf0_button_end
	ldi	button, BUTTON_LONG_TIME
	brne	ovf0_button_end
	ori	flags, 1<<fButtonLong
	rjmp	ovf0_button_end
ovf0_button_released:
	mov	inttmp, button
	clr	button
	cpi	inttmp, BUTTON_SHORT_TIME
	brlo	ovf0_button_end
	cpi	inttmp, BUTTON_LONG_TIME
	brsh	ovf0_button_end
	ori	flags, 1<<fButtonShort
ovf0_button_end:

	;decrement delay and update LEDs (~100 Hz) - max. 197 cycles
	tst	delay
	breq	ovf0_update
	dec	delay
	brne	ovf0_update_end
ovf0_update:
	sbrs	flags, fUpdateLEDs
	rjmp	ovf0_update_end
	andi	flags, ~(1<<fUpdateLEDs)
	mov	delay, new_delay
	ldi	ZH, HIGH(RAM_FrameShadow+34)
	ldi	ZL, LOW(RAM_FrameShadow+34)
	ldi	intcnt, 34/2
ovf0_update_loop:
	ld	inttmp, -Z
	std	Z+(RAM_FrameLED - RAM_FrameShadow), inttmp
	ld	inttmp, -Z
	std	Z+(RAM_FrameLED - RAM_FrameShadow), inttmp
	dec	intcnt
	brne	ovf0_update_loop
ovf0_update_end:

ovf0_end:
	;84 cycles from here to end of interrupt (including 'reti')
	
	;calculate new values for LEDs (output in next interrupt)
	ldi	ZH, HIGH(RAM_FrameLED) ;RAM address = FrameLED + 6 * mux
	ldi	ZL, LOW(RAM_FrameLED)
	sbrc	mux, 1
	adiw	ZH:ZL, 1*6
	sbrc	mux, 2
	adiw	ZH:ZL, 2*6
	sbrc	mux, 3
	adiw	ZH:ZL, 3*6
	sbrc	mux, 4
	adiw	ZH:ZL, 4*6
	sbrc	mux, 5
	adiw	ZH:ZL, 5*6
	clr	leds
	ldi	intcnt, 0x01
ovf0_led_loop:
	ld	inttmp, Z+
	cp	pwm, inttmp
	brsh	PC+2
	or	leds, intcnt
	lsl	intcnt
	sbrs	intcnt, 6
	rjmp	ovf0_led_loop
	
	;load timer value for next PWM interval (for next interrupt)
	mov	ZL, pwm
	ldi	ZH, 0
	subi	ZL, LOW(-pwm_table*2)
	sbci	ZH, HIGH(-pwm_table*2)
	lpm	pwm_timer, Z
	
	;return
	pop	ZH
	pop	ZL
	out	SREG, sreg_backup
	reti
	
;--------------------

ovf1:
	;timer 1 overflow interrupt: load PWM with next audio sample
	in	sreg_backup, SREG
	
	;read byte from flash and output to PWM registers
	lpm	inttmp, Z+
	cpi	inttmp, 0 ;no more data?
	breq	ovf1_stop
	_out	OCR1AL, inttmp ;set new PWM value
	neg	inttmp
	_out	OCR1BL, inttmp
	
ovf1_end:
	;return
	out	SREG, sreg_backup
	reti
	
ovf1_stop:
	;playback finished
	_in	inttmp, TIMSK1
	andi	inttmp, ~(1<<TOIE1)
	_out	TIMSK1, inttmp
	rjmp	ovf1_end
	
;===============================================================================

;animation command macros

.macro FILL ;FILL led, n, brt
  .if ((@0 < 1) || (@0 > 34))
    .error "FILL: LED number out of range (1~34)."
  .endif
  .if ((@1 < 1) || (@1 > 34))
    .error "FILL: stretch length out of range (1~34)."
  .endif
  .if ((@2 < 0) || (@2 > 7))
    .error "FILL: Brightness out of range (0~7)."
  .endif
  .db ((@0 - 1) << 1) | (@1 >> 5), LOW(@1 << 3) | @2
.endmacro

.macro LED ;LED led, brt
  FILL @0, 1, @1
.endmacro

.macro CLEAR ;CLEAR
  FILL 1, 34, 0
.endmacro

.macro ROTATE ;ROTATE n
  .if ((@0 >= 1) && (@0 <= 33))
    .db 0x80, @0
  .else ;.elseif doesn't work: "Found no label/variable/constant named" (sic)
    .if ((@0 >= -33) && (@0 <= -1))
      .db 0x80, 34 + @0
    .else
      .error "ROTATE: value out of range (-33~-1 or 1~33)."
    .endif
  .endif
.endmacro

.macro FADE ;FADE n
  .if (((@0 >= 1) && (@0 <= 7)) || ((@0 >= -7) && ( @0 <= -1)))
    .db 0x81, @0
  .else
    .error "FADE: value out of range (-7~-1 or 1~7)."
  .endif
.endmacro

.macro DELAY ;DELAY n
  .if ((@0 < 10) || (@0 > 255 * 10))
    .error "DELAY: value out of range (10~2550)."
  .endif
  .db 0x82, LOW((@0 + 5) / 10)
.endmacro

.macro REPEAT ;REPEAT n
  .if ((@0 < 1) || (@0 > 255))
    .error "REPEAT: value out of range (1~255)."
  .endif
  .db 0x83, @0
.endmacro

.macro REPEAT_END ;REPEAT_END
  .db 0x83, 0
.endmacro

.macro ANI_END ;ANI_END
  .db 0x84, 0
.endmacro

;===============================================================================

pwm_table:
	.db -(PWM_1 -   0  ), -(PWM_2 - PWM_1)
	.db -(PWM_3 - PWM_2), -(PWM_4 - PWM_3)
	.db -(PWM_5 - PWM_4), -(PWM_6 - PWM_5)
	.db -(PWM_7 - PWM_6), 0

animations:
	.include "animations/ani_list.asm"

audio:
	.include "audio/quack32500_u8.asm"
	
;===============================================================================

reset:
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;init ports
.ifdef (PORTA)
	ldi	temp, 0x03 ;ADC7:6 (ATmega) = PA1:0 (ATtiny88): enable pull-ups
	out	PORTA, temp
	;GND(pin3) (ATmega) = PA2 (ATtiny88): hardwired to GND
	;VCC(pin6) (ATmega) = PA3 (ATtiny88): hardwired to VCC
.endif
	ldi	temp, 0x09 ;enable pull-up for push button and always-on inputs
	out	PORTB, temp
	ldi	temp, 0xF6
	out	DDRB, temp
	ldi	temp, 0x00
	out	PORTC, temp
	ldi	temp, 0x80|0x3F ;AREF (ATmega) = PC7 (ATtiny88): set to output
	out	DDRC, temp
	ldi	temp, 0x80 ;prototype: amplifier off
	out	PORTD, temp
	ldi	temp, 0xFF
	out	DDRD, temp

	;reset watchdog
	wdr
	
	;init watchdog timer
	ldi	temp, 1<<WDCE | 1<<WDE ;(timed sequence for access to WDTCR)
	_out	WDTCSR, temp
	ldi	temp, 1<<WDIE | 1<<WDP2 ;interrupt mode, 256k cycles (~0.25 s)
	_out	WDTCSR, temp
	
;	;check for watchdog reset
;	in	temp, MCUSR
;	ldi	temp2, 0
;	out	MCUSR, temp2
;	sbrs	temp, WDRF
;	rjmp	wdr_end
;	sbrc	flags, fExpectWdr
;	rjmp	wdr_wakeup
;wdr_end:
	;init stack pointer
	ldi	temp, HIGH(RAMEND)
	out	SPH, temp
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;init RAM
	ldi	ZL, LOW(RAM_FrameShadow)
	ldi	ZH, HIGH(RAM_FrameShadow)
	ldi	temp, 0
	ldi	count, 36 ;virtual matrix size is 6x6 LEDs
init_ram:
	std	Z+(RAM_FrameLED - RAM_FrameShadow), temp
	st	Z+, temp
	dec	count
	brne	init_ram
	
	;init registers
	ldi	pwm, 0
	ldi	mux, 0x20
	ldi	temp, 0xFF
	mov	pwm_timer, temp
	clr	leds
	clr	delay
	clr	button
	clr	flags
	ldi	ZL, LOW(animations*2)
	ldi	ZH, HIGH(animations*2)
	
	;enable interrupts
	sei
	
	;go to standby
	sbic	PINB, 3 ;always-on mode?
	rjmp	power_off
	rjmp	power_on
	
;-------------------------------------------------------------------------------
	
ani_start:
	ldi	ZL, LOW(animations*2)
	ldi	ZH, HIGH(animations*2)
ani_loop:
	lpm	temp, Z+
	cpi	temp, 0x80 ;1000.0000 - rotate
	breq	ani_rotate
	brlo	ani_fill ;0xxx.xxxx - fill LEDs with value
	cpi	temp, 0x81 ;1000.0001 - fade
	breq	ani_fade
	cpi	temp, 0x82 ;1000.0010 - delay
	breq	ani_delay
	cpi	temp, 0x83 ;1000.0011 - repeat, repeat end
	brne	PC+2
	rjmp	ani_repeat
	cpi	temp, 0x84 ;1000.0100 - end of animations
	breq	ani_start
	;1000.0100 ~ 1111.1111 reserved for future extensions
	rjmp	ani_loop
	
	
ani_delay:
	;delay: new_delay * 10 ms
	lpm	new_delay, Z+
	ori	flags, 1<<fUpdateLEDs
ani_delay_wait:
	sbrc	flags, fButtonShort
	rcall	sound
	sbrs	flags, fButtonLong
	rjmp	ani_delay_wait_end
	sbic	PINB, 3 ;always-on mode?
	rjmp	power_off
power_off_return:
ani_delay_wait_end:
	sbrc	flags, fUpdateLEDs
	rjmp	ani_delay_wait
	rjmp	ani_loop
	
ani_fade:
	;fade
	lpm	temp, Z+
	ldi	YL, LOW(RAM_FrameShadow)
	ldi	YH, HIGH(RAM_FrameShadow)
	ldi	count, 34
ani_fade_loop:
	ld	temp2, Y
	add	temp2, temp
	sbrc	temp2, 7 ;negative?
	ldi	temp2, 0
	sbrc	temp2, 3 ;larger than 7 (max. 7+7 = 14)
	ldi	temp2, 7
	st	Y+, temp2
	dec	count
	brne	ani_fade_loop
	rjmp	ani_loop
	
ani_fill:
	;0aaa.aaal, llll.lvvv - fill l LEDs, starting at address a, with value v
	lpm	count, Z+
	ldi	temp2, 0
	lsr	temp
	ror	count
	ror	temp2
	lsr	count
	ror	temp2
	lsr	count
	ror	temp2
	lsr	temp2
	swap	temp2
	mov	YL, temp
	ldi	YH, 0
	subi	YL, LOW(-RAM_FrameShadow)
	sbci	YH, HIGH(-RAM_FrameShadow)
ani_fill_loop:
	cpi	YL, LOW(RAM_FrameShadow+34) ;wrap around for address
	brne	PC+2
	sbiw	YH:YL, 34
	st	Y+, temp2 ;set LED
	dec	count
	brne	ani_fill_loop
	rjmp	ani_loop
	
ani_rotate:
	;rotate
	;copy LED values to backup framebuffer at shifted address
	lpm	temp, Z+
	ldi	XL, LOW(RAM_FrameShadow)
	ldi	XH, HIGH(RAM_FrameShadow)
	mov	YL, temp
	ldi	YH, 0
	subi	YL, LOW(-RAM_FrameBackup)
	sbci	YH, HIGH(-RAM_FrameBackup)
	ldi	count, 34
ani_rotate_loop1:
	cpi	YL, LOW(RAM_FrameBackup+34)
	brne	PC+2
	sbiw	YH:YL, 34
	ld	temp, X+
	st	Y+, temp
	dec	count
	brne	ani_rotate_loop1
	;copy shifted values back to displayed framebuffer
	ldi	YL, LOW(RAM_FrameBackup+34)
	ldi	YH, HIGH(RAM_FrameBackup+34)
	ldi	count, 34
ani_rotate_loop2:
	ld	temp, -Y
	std	Y+(RAM_FrameShadow - RAM_FrameBackup), temp
	dec	count
	brne	ani_rotate_loop2
	rjmp	ani_loop
	
ani_repeat:
	;repeat
	lpm	temp, Z+
	cpi	temp, 0
	breq	ani_repeat_end
	push	repeat ;store repeat counter for possible surrounding loop
	push	ZL ;store start-of-repeat-block address
	push	ZH
	mov	repeat, temp
	rjmp	ani_loop
	
ani_repeat_end:
	;repeat end
	dec	repeat
	breq	ani_repeat_complete
	pop	ZH ;read start-of-repeat-block address from stack
	pop	ZL
	push	ZL
	push	ZH
	rjmp	ani_loop
ani_repeat_complete:
	pop	temp ;remove start-of-repeat-block address from stack
	pop	temp
	pop	repeat ;restore repeat counter for possible surrounding loop
	rjmp	ani_loop
	
;--------------------

power_off:
	;disable PWM/mux timer
	ldi	temp, 0 ;stop timer
	out	TCCR0B, temp
	ldi	temp, 0 ;disable timer overflow interrupt
	_out	TIMSK0, temp
	_out	TIMSK1, temp
	
	;LEDs off
	ldi	temp, 0x00
	out	PORTC, temp
	ldi	temp, 0x80
	out	PORTD, temp
	
	;backup stackpointer
	in	spl_backup, SPL
	in	sph_backup, SPH
	
	ori	flags, 1<<fExpectWdr
power_off_sleep1:
	ldi	wakeup_count, 4
power_off_sleep2:
	;sleep
	ldi	temp, 1<<SE | 1<<SM1
	out	SMCR, temp
	sleep
;wdr_wakeup:
	
	;check if button is being pressed
	sbic	PINB, 0
	rjmp	power_off_sleep1
	dec	wakeup_count
	brne	power_off_sleep2
	
	;disable sleeping
	ldi	temp, 0
	out	SMCR, temp
	andi	flags, ~(1<<fExpectWdr)
	
	;restore stackpointer
	out	SPL, spl_backup
	out	SPH, sph_backup
	
power_on:
	;re-enable PWM/mux timer
	ldi	temp, 0x03 ;Clk/64
	_out	TCCR0B, temp
	ldi	temp, 1<<TOIE0
	_out	TIMSK0, temp
	
	ldi	button, BUTTON_LONG_TIME + 1
	andi	flags, ~(1<<fButtonShort | 1<<fButtonLong)
	rjmp	power_off_return

;--------------------

sound:

	;backup registers
	push	delay
	push	ZL
	push	ZH
	
	;enable audio PWM on OC1A/OC1B, fast PWM mode, non-inv, 8 bit
	ldi	temp, 0
	_out	TCNT1H, temp
	_out	TCNT1L, temp
	ldi	temp, 0
	_out	OCR1AH, temp
	ldi	temp, 0x80 ;idle: 50% duty cycle
	_out	OCR1AL, temp
	ldi	temp, 0
	_out	OCR1BH, temp
	ldi	temp, 0x80 ;idle: 50% duty cycle
	_out	OCR1BL, temp
	ldi	temp, 1<<WGM10 | 1<<COM1A1 | 1<<COM1B1 ;fast PWM, non-inv.
	_out	TCCR1A, temp
	ldi	temp, 1<<WGM12 | 0x01 ;fast PWM, Clk/1
	_out	TCCR1B, temp
	
	;enable amplifier (prototype)
	cbi	PORTD, 7
	
	;stop further display updates
	andi	flags, ~(1<<fUpdateLEDs)
	clr	delay
	
	;backup framebuffer
	ldi	ZH, HIGH(RAM_FrameBackup)
	ldi	ZL, LOW(RAM_FrameBackup)
	ldi	count, 34
sound_leds_backup:
	ldd	temp, Z+(RAM_FrameShadow - RAM_FrameBackup)
	st	Z+, temp
	dec	count
	brne	sound_leds_backup
	
	;fade all LEDs on
fade_on:
	ldi	count, 1
fade_on_loop1:
	ldi	ZH, HIGH(RAM_FrameShadow)
	ldi	ZL, LOW(RAM_FrameShadow)
	ldi	temp2, 34
fade_on_loop2:
	ld	temp, Z
	cp	temp, count
	brsh	PC+2
	st	Z, count
	adiw	ZH:ZL, 1
	dec	temp2
	brne	fade_on_loop2
	ldi	temp, 2 ;20 ms delay
	mov	new_delay, temp
	ori	flags, 1<<fUpdateLEDs
fade_on_wait:
	sbrc	flags, fUpdateLEDs
	rjmp	fade_on_wait
	inc	count
	cpi	count, 8
	brlo	fade_on_loop1
	
	;delay until LEDs updated
amp_on_delay:
	tst	delay
	brne	amp_on_delay
	
	;start playback
	ldi	ZH, HIGH(audio*2)
	ldi	ZL, LOW(audio*2)
	ldi	temp, 1<<TOV1 ;clear timer overflow interrupt flag
	out	TIFR1, temp
	_in	temp, TIMSK1
	ori	temp, 1<<TOIE1 ;enable timer overflow interrupt
	_out	TIMSK1, temp
	
	;wait until playback finished
sound_wait:
	_in	temp, TIMSK1
	sbrc	temp, TOIE1
	rjmp	sound_wait
	
	;disable amplifier (prototype)
	sbi	PORTD, 7
	
	;fade all LEDs back to original values
fade_off:
	ldi	count, 6
fade_off_loop1:
	ldi	ZH, HIGH(RAM_FrameBackup+34)
	ldi	ZL, LOW(RAM_FrameBackup+34)
	ldi	temp2, 34
fade_off_loop2:
	ld	temp, -Z
	cp	count, temp
	brlo	PC+2
	std	Z+(RAM_FrameShadow - RAM_FrameBackup), count
	dec	temp2
	brne	fade_off_loop2
	ldi	temp, 2 ;20 ms delay
	mov	new_delay, temp
	ori	flags, 1<<fUpdateLEDs
fade_off_wait:
	sbrc	flags, fUpdateLEDs
	rjmp	fade_off_wait
	subi	count, 1
	brcc	fade_off_loop1
	
	;disable audio PWM
	ldi	temp, 0
	_out	TCCR1B, temp
	_out	TCCR1A, temp
	
	;restore registers
	pop	ZH
	pop	ZL
	pop	delay

	;return
	andi	flags, ~(1<<fButtonShort)
	ori	flags, 1<<fUpdateLEDs
	ret
	